#include <iostream>
#include <memory.h>
#include <fltk/math.h>
#include <xtiffio.h>
#include <geotiffio.h>
#include "Map.h"

double laserRange = 8.0f;
double l0 = 0.0f;
double locc = .5f;
double lfree = -.1f;

Map::Map()
{
    bFirst = true;
    w = MY_MAP_WIDTH;
    h = MY_MAP_HEIGHT;
    //map = (double **)malloc(sizeof(double)*w*h);
    //memset((void *)map,0,sizeof(double)*w*h);
    for (int x=0;x<w;x++)
        for (int y=0;y<h;y++)
            map[x][y] = 0.0f;
}

Map::~Map()
{
    //free(map);
}

void Map::UpdateGridSquare(int x,int y,double value)
{
    map[x][y] += value;
}

void Map::AddTrace(double startx, double starty, double MeasuredDist, double theta,int addEnd)
{
    double overflow, slope; // Used for actually tracing the line
    int x, y, incX, incY, endx, endy;
    int xedge, yedge;       // Used in computing the midpoint. Recompensates for which edge of the square the line entered from
    double dx, dy;
    double distance, error;
    double secant, cosecant;   // precomputed for speed

    // Precomute a few numbers for speed.
    secant = 1.0/fabs(cos(theta));
    cosecant = 1.0/fabs(sin(theta));

    // This allows the user to limit the effective range of the sensor
    distance = MIN(MeasuredDist,laserRange);

    // Mark the final endpoint of the line trace, so that we know when to stop.
    // We keep the endpoint as both float and int.
    dx = (startx + (cos(theta) * distance));
    dy = (starty + (sin(theta) * distance));
    endx = (int) (dx);
    endy = (int) (dy);

    // Decide which x and y directions the line is travelling.
    // inc tells us which way to increment x and y. edge indicates whether we are computing
    // distance from the near or far edge.
    if (startx > dx) {
        incX = -1;
        xedge = 1;
    }
    else {
        incX = 1;
        xedge = 0;
    }

    if (starty > dy) {
        incY = -1;
        yedge = 1;
    }
    else {
        incY = 1;
        yedge = 0;
    }

    // Figure out whether primary motion is in the x or y direction. 
    // The two sections of code look nearly identical, with x and y reversed.
    if (fabs(startx - dx) > fabs(starty - dy)) {
        // The given starting point is non-integer. The line therefore starts at some point partially set in to the starting
        // square. Overflow starts at this offcenter amount, in order to make steps in the y direction at the right places.
        y = (int) (starty);
        overflow =  starty - y;
        // We always use overflow as a decreasing number- therefore positive y motion needs to 
        // adjust the overflow value accordingly.
        if (incY == 1)
            overflow = 1.0 - overflow;
        // Compute the effective slope of the line.
        slope = fabs(tan(theta));

        // The first square is a delicate thing, as we aren't doing a full square traversal in
        // either direction. So we figure out this strange portion of a step so that we can then
        // work off of the axes later. 
        // NOTE: we aren't computing the probability of this first step. Its a technical issue for
        // simplicity, and the odds of the sensor sitting on top of a solid object are sufficiently 
        // close to zero to ignore this tiny portion of a step. 
        error = fabs(((int)(startx)+incX+xedge)-startx);
        overflow = overflow - (slope*error);
        // The first step is actually in the y direction, due to the proximity of starty to the y axis. 
        if (overflow < 0.0) {
            y = y + incY;
            overflow = overflow + 1.0;
        }

        // Now we can start the actual line trace.
        for (x = (int) (startx) + incX; x != endx; x = x + incX) {
            overflow = overflow - slope;

            // Compute the distance travelled in this square
            if (overflow < 0.0)
                distance = (overflow+slope)*cosecant;
            else
                distance = fabs(slope)*cosecant;
            // Update every grid square we cross as empty...
            UpdateGridSquare(x,y,lfree);

            // ...including the overlap in the minor direction
            if (overflow < 0) {
                y = y + incY;
                distance = -overflow*cosecant;
                overflow = overflow + 1.0;
                UpdateGridSquare(x,y,lfree);
            }
        }

        // Update the last grid square seen as having a hit.
        if (addEnd) {
            if (incX < 0)
                distance = fabs((x+1) - dx)*secant;
            else
                distance = fabs(dx - x)*secant;
            UpdateGridSquare(endx,endy,locc);
        }

    }

    // This is the same as the previous block of code, with x and y reversed.
    else {
        x = (int) (startx);
        overflow = startx - x;
        if (incX == 1)
            overflow = 1.0 - overflow;
        slope = 1.0/fabs(tan(theta));

        // (See corresponding comments in the previous half of this function)
        error = fabs(((int)(starty)+incY+yedge)-starty);
        overflow = overflow - (error*slope);
        if (overflow < 0.0) {
            x = x + incX;
            overflow = overflow + 1.0;
        }

        for (y = (int) (starty) + incY; y != endy; y = y + incY) {
            overflow = overflow - slope;
            if (overflow < 0)
                distance = (overflow+slope)*secant;
            else
                distance = fabs(slope)*secant;

            UpdateGridSquare(x,y,lfree);

            if (overflow < 0.0) {
                x = x + incX;
                distance = -overflow*secant;
                overflow = overflow + 1.0;
                UpdateGridSquare(x,y,lfree);
            }
        }

        if (addEnd) {
            if (incY < 0)
                distance = fabs(((y+1) - dy)/sin(theta));
            else
                distance = fabs((dy - y)/sin(theta));
            UpdateGridSquare(endx,endy,locc);
        }
    }
}

void Map::UpdateOccupancyGrid(double x,double y,double theta,double* laser)
{
    if (bFirst)
    {
        initialPos[0] = x;
        initialPos[1] = y;
        initialPos[2] = theta;
        bFirst = false;
    }
    sleep(1);
    /*
    for (int i=0;i<180;i++)
    {
        double centerx = (double)w/2.0f;
        double centery = (double)h/2.0f;
        AddTrace(   centerx-initialPos[0]+x,
                    centery-initialPos[1]+y,
                    -initialPos[2]+theta,laser[i],(laser[i] < laserRange));
    }*/
}
bool Map::ExportGeoTiff(char* s)
{
    TIFF* tiff = NULL;
    GTIF* gtif = NULL;

    tiff = XTIFFOpen(s,"w");
    if (tiff==NULL)
        return false;

    gtif = GTIFNew(tiff);
    if (gtif==NULL)
        return false;

    double tiepoints[6]={0.0,0.0,0.0,130.0,32.0,0.0};
    double pixscale[3]={1.0,1.0,0.0};

    //cout << "creating directory: " << TIFFCreateDirectory(tiff) << endl;
    TIFFSetField(tiff,TIFFTAG_IMAGEWIDTH,w);
    TIFFSetField(tiff,TIFFTAG_IMAGELENGTH,h);
    TIFFSetField(tiff,TIFFTAG_BITSPERSAMPLE,8);
    //TIFFSetField(tiff,TIFFTAG_COMPRESSION,COMPRESSION_PACKBITS);
    TIFFSetField(tiff,TIFFTAG_COMPRESSION,COMPRESSION_DEFLATE);
    TIFFSetField(tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
    TIFFSetField(tiff,TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_MINISBLACK);
    //TIFFSetField(tiff,TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_RGB);
    //TIFFSetField(tiff,TIFFTAG_SAMPLESPERPIXEL,3);
    //TIFFSetField(tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_SEPARATE);
    //TIFFSetField(tiff,TIFFTAG_SAMPLESPERPIXEL,2);
    //TIFFSetField(tiff,TIFFTAG_ROWSPERSTRIP,20L);
    //TIFFSetField(tiff,TIFFTAG_GEOTIEPOINTS,6,tiepoints);
    //TIFFSetField(tiff,TIFFTAG_GEOPIXELSCALE,3,pixscale);

    //*
    GTIFKeySet(gtif,GTModelTypeGeoKey,TYPE_SHORT,1,ModelGeographic);
    GTIFKeySet(gtif, GTModelTypeGeoKey, TYPE_SHORT, 1, ModelGeographic);
	GTIFKeySet(gtif, GTRasterTypeGeoKey, TYPE_SHORT, 1, RasterPixelIsArea);
	GTIFKeySet(gtif, GTCitationGeoKey, TYPE_ASCII, 0, "Probablity Occupancy Map");
	GTIFKeySet(gtif, GeographicTypeGeoKey, TYPE_SHORT,  1, KvUserDefined);
	//GTIFKeySet(gtif, GeogCitationGeoKey, TYPE_ASCII, 0, "Everest Ellipsoid Used.");
	GTIFKeySet(gtif, GeogAngularUnitsGeoKey, TYPE_SHORT,  1, Angular_Degree);
	GTIFKeySet(gtif, GeogLinearUnitsGeoKey, TYPE_SHORT,  1, Linear_Meter);
	GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT,     1, KvUserDefined);
	//GTIFKeySet(gtif, GeogEllipsoidGeoKey, TYPE_SHORT,     1, Ellipse_Everest_1830_1967_Definition);
	//GTIFKeySet(gtif, GeogSemiMajorAxisGeoKey, TYPE_DOUBLE, 1, (double)6377298.556);
	//GTIFKeySet(gtif, GeogInvFlatteningGeoKey, TYPE_DOUBLE, 1, (double)300.8017);

    //*/

    char buffer[w];

    for (int y=0;y<h;y++)
    {
        for (int x=0;x<w;x++)
        {
            int color = 255;
            if      (map[x][y] > .5f)   color = 0;
            else if (map[x][y] < -.5f)  color = 255;
            else
                color = (int)(255.0f - round( ((map[x][y] + .5f) * 255.0f)));

            buffer[x] = color;
        }
        if (!TIFFWriteScanline(tiff,buffer,y,1))
        {
            TIFFError("writeGeoTiff","failure in WriteScanline\n");
            return false;
        }
    }
    
    GTIFFree(gtif);
    XTIFFClose(tiff);
    
    return true;
}
